using System;
using System.Collections.Concurrent;
using System.Net;
using System.Net.Sockets;
using System.Net.NetworkInformation;

namespace RoboidControl {

    /// <summary>
    /// A participant using UDP communication
    /// </summary>
    /// A local participant is the local device which can communicate with
    /// other participants It manages all local things and communcation with other
    /// participants. Each application has a local participant which is usually
    /// explicit in the code. An participant can be isolated. In that case it is
    /// standalong and does not communicate with other participants.
    ///
    /// It is possible to work with an hidden participant by creating things without
    /// specifying a participant in the constructor. In that case an hidden isolated
    /// participant is created which can be obtained using
    /// RoboidControl::IsolatedParticipant::Isolated().
    /// @sa RoboidControl::Thing::Thing()
    public class ParticipantUDP : Participant {

        #region Init

        /// <summary>
        /// Create a participant without connecting to a site
        /// </summary>
        /// <param name="port">The port number on which to communicate</param>
        /// These participant typically broadcast Participant messages to let site
        /// servers on the local network know their presence. Alternatively they can
        /// broadcast information which can be used directly by other participants.
        public ParticipantUDP(int port = 0) : base("127.0.0.1", port) {
            if (this.port == 0)
                this.isIsolated = true;
            Participant.AddParticipant(this);
            Participant.ReplaceLocalParticipant(this);
        }
        /// <summary>
        ///  Create a participant which will try to connect to a site.
        /// </summary>
        /// <param name="ipAddress">The ip address of the site server</param>
        /// <param name="port">The port number of the site server</param>
        /// <param name="localPort">The port used by the local participant</param>
        public ParticipantUDP(string ipAddress, int port = 7681, int localPort = 7681) : base("127.0.0.1", localPort) {
            this.ipAddress = ipAddress;
            this.port = port;

            // Determine local IP address
            IPAddress localIpAddress = null;
            IPAddress subnetMask = null;
            using (Socket socket = new(AddressFamily.InterNetwork, SocketType.Dgram, 0)) {
                // Connect to a remote endpoint — we won't actually send anything
                socket.Connect("1.1.1.1", 65530);
                if (socket.LocalEndPoint is IPEndPoint endPoint) {
                    localIpAddress = endPoint.Address;
                    this.ipAddress = localIpAddress.ToString();
                }
            }
            // Find subnet mask
            foreach (NetworkInterface nwInterface in NetworkInterface.GetAllNetworkInterfaces()) {
                if (nwInterface.OperationalStatus != OperationalStatus.Up)
                    continue;

                foreach (UnicastIPAddressInformation unicastAddress in nwInterface.GetIPProperties().UnicastAddresses) {
                    if (unicastAddress.Address.AddressFamily == AddressFamily.InterNetwork && unicastAddress.Address.Equals(localIpAddress)
                    ) {
                        subnetMask = unicastAddress.IPv4Mask;
                    }
                }
            }
            if (localIpAddress != null && subnetMask != null)
                GetBroadcastAddress(localIpAddress, subnetMask);

            this.endPoint = new IPEndPoint(IPAddress.Any, localPort);
            this.udpClient = new UdpClient(localPort);
            this.udpClient.BeginReceive(new AsyncCallback(result => ReceiveUDP(result)), null);

            if (this.port == 0)
                this.isIsolated = true;
            else
                this.remoteSite = new ParticipantUDP(ipAddress, port, this);

            Participant.AddParticipant(this);
            Participant.ReplaceLocalParticipant(this);
        }

        public ParticipantUDP(string ipAddress, int port, ParticipantUDP localParticipant) : base(ipAddress, port, localParticipant) {
            this.ipAddress = ipAddress;
            this.port = port;

            if (localParticipant != null)
                this.udpClient = localParticipant.udpClient;

        }

        #endregion Init

        #region Properties

        /// <summary>
        /// The name of the participant
        /// </summary>
        public string name = "ParticipantUDP";

        /// <summary>
        /// The Ip Address of a participant. When the participant is local, this contains 0.0.0.0
        /// </summary>
        /// <remarks>This does not belong here, it should move to ParticipantUDP or something like that in the future
        public string ipAddress = "0.0.0.0";
        /// <summary>
        /// The port number for UDP communication with the participant. This is 0 for isolated participants.
        /// </summary>
        /// <remarks>This does not belong here, it should move to ParticipantUDP or something like that in the future
        public int port = 0;

        /// <summary>
        /// The udpClient for this participant
        /// </summary>
        /// <remarks>This does not belong here, it should move to ParticipantUDP or something like that in the future
        public UdpClient udpClient = null;

        /// <summary>
        /// True if the participant is running isolated.
        /// </summary>
        /// Isolated participants do not communicate with other participants
        public bool isIsolated = false;

        /// <summary>
        /// The remote site when this participant is connected to a site
        /// </summary>
        public Participant remoteSite = null;

        /// <summary>
        /// The interval in milliseconds for publishing (broadcasting) data on the local network
        /// </summary>
        public ulong publishIntervalMS = 3000; // = 3 seconds
        public ulong sendUpdateIntervalMS = 100; // = 0.1 seconds for object updates

        public IPEndPoint endPoint = null;
        public string broadcastIpAddress = "255.255.255.255";

        public readonly ConcurrentQueue<IMessage> messageQueue = new ConcurrentQueue<IMessage>();

        protected void GetBroadcastAddress(IPAddress ip, IPAddress subnetMask) {
            byte[] ipBytes = ip.GetAddressBytes();
            byte[] maskBytes = subnetMask.GetAddressBytes();

            if (ipBytes.Length != maskBytes.Length)
                throw new ArgumentException("IP address and subnet mask lengths do not match");

            byte[] broadcastBytes = new byte[ipBytes.Length];
            for (int i = 0; i < ipBytes.Length; i++) {
                broadcastBytes[i] = (byte)(ipBytes[i] | (~maskBytes[i]));
            }

            IPAddress broadcastAddress = new(broadcastBytes);
            this.broadcastIpAddress = broadcastAddress.ToString();

            // Console.WriteLine($"Subnet mask: {subnetMask.ToString()}");
            // Console.WriteLine($"Broadcast address: {this.broadcastIpAddress}");
        }

        #endregion Properties

        #region Update

        public override void Update() {
            ulong currentTimeMS = Thing.GetTimeMs();

            if (this.isIsolated == false && this.isRemote == false) {
                if (this.publishIntervalMS > 0 && currentTimeMS > this.nextPublishMe) {
                    ParticipantMsg msg = new(this.networkId);
                    if (this.remoteSite == null)
                        this.Publish(msg);
                    else
                        this.remoteSite.Send(msg);

                    this.nextPublishMe = currentTimeMS + this.publishIntervalMS;
                }
            }

            UpdateMyThings(currentTimeMS);
            if (this.isRemote == false)
                UpdateOtherThings(currentTimeMS);
        }

        protected ulong nextPublishMe = 0;
        protected ulong nextSendUpdate = 0;

        protected virtual void UpdateMyThings(ulong currentTimeMS) {
            foreach (Thing thing in this.things) {
                if (thing == null)
                    continue;

                // if (thing.hierarchyChanged && !(this.isIsolated || this.networkId == 0)) {
                //     ThingMsg thingMsg = new(this.networkId, thing);
                //     // this.Send(this.remoteSite, thingMsg);
                //     this.remoteSite.Send(thingMsg);
                // }

                // Why don't we do recursive?
                // Because when a thing creates a thing in the update,
                // that new thing is not sent out (because of hierarchyChanged)
                // before it is updated itself: it is immediatedly updated!
                thing.Update(false);
                if (!(this.isIsolated || this.networkId == 0)) {
                    if (thing.terminate) {
                        DestroyMsg destroyMsg = new(this.networkId, thing);
                        // this.Send(this.remoteSite, destroyMsg);
                        this.remoteSite.Send(destroyMsg);
                    }
                    // else {
                    //     // Send to remote site
                    //     // PoseMsg poseMsg = new(thing.owner.networkId, thing);
                    //     // this.Send(this.remoteSite, poseMsg);
                    //     // BinaryMsg binaryMsg = new(thing.owner.networkId, thing);
                    //     // this.Send(this.remoteSite, binaryMsg);
                    //     this.remoteSite.Send(new PoseMsg(thing.owner.networkId, thing));
                    //     this.remoteSite.Send(new BinaryMsg(thing.owner.networkId, thing));
                    // }
                }
                if (thing.terminate)
                    this.Remove(thing);
            }
        }

        protected virtual void UpdateOtherThings(ulong currentTimeMS) {
            for (int ownerIx = 0; ownerIx < Participant.participants.Count; ownerIx++) {
                Participant owner = Participant.participants[ownerIx];
                if (owner == null || owner == this)
                    continue;

                owner.Update();
                // if (this.isIsolated)
                //     continue;

                if (currentTimeMS > this.nextSendUpdate) {
                    for (int thingIx = 0; thingIx < owner.things.Count; thingIx++) {
                        Thing thing = owner.things[thingIx];
                        if (thing == null)
                            continue;

                        foreach (Participant participant in Participant.participants) {
                            if (participant == this || participant == owner)
                                continue;

                            Console.WriteLine($"[{thing.owner.networkId}/{thing.id}] Pose ---> {participant.networkId}");
                            PoseMsg poseMsg = new(thing.owner.networkId, thing, true);
                            participant.Send(poseMsg);
                        }
                    }
                }
            }
            if (currentTimeMS > this.nextSendUpdate)
                this.nextSendUpdate = currentTimeMS + this.sendUpdateIntervalMS;
        }

        #endregion Update

        #region Send

        public override bool Send(IMessage msg) {
            int bufferSize = msg.Serialize(ref this.buffer);
            if (bufferSize <= 0)
                return true;

            IPEndPoint participantEndpoint = new IPEndPoint(IPAddress.Parse(this.ipAddress), this.port);
            // Console.WriteLine($"msg to remote participant {participantEndpoint.Address.ToString()} {participantEndpoint.Port}");
            if (udpClient != null) {
                //Console.WriteLine("sending...");
                this.udpClient?.Send(this.buffer, bufferSize, participantEndpoint);
            }
            return true;
        }

        public void PublishThingInfo(Thing thing) {
            // Console.WriteLine("Publish thing info");
            this.Publish(new ThingMsg(this.networkId, thing));
            this.Publish(new NameMsg(this.networkId, thing));
            this.Publish(new ModelUrlMsg(this.networkId, thing));
            this.Publish(new PoseMsg(this.networkId, thing));
            this.Publish(new BinaryMsg(this.networkId, thing));
        }

        public bool Publish(IMessage msg) {
            int bufferSize = msg.Serialize(ref this.buffer);
            if (bufferSize <= 0)
                return true;

            Console.WriteLine($"publish to {broadcastIpAddress.ToString()} {this.port}");
            this.udpClient?.Send(this.buffer, bufferSize, this.broadcastIpAddress, this.port);
            return true;
        }

        #endregion

        #region Receive

        protected void ReceiveUDP(IAsyncResult result) {
            try {
                if (this.udpClient == null)
                    return;

                byte[] data = { };
                try {
                    data = this.udpClient.EndReceive(result, ref endPoint);
                }
                catch (SocketException e) {
                    if (e.ErrorCode != 10054)
                        throw;
#if DEBUG                        
                    else
                        Console.WriteLine($"No receiver found for the previously sent UDP packet");
#endif                        
                }
                // This does not yet take multi-packet messages into account!
                if (endPoint == null)
                    return;

                if (data.Length > 0) {
                    // We can receive our own publish (broadcast) packages. How do we recognize them????
                    // It is hard to determine our source port
                    string ipAddress = endPoint.Address.ToString();
                    if (ipAddress != this.ipAddress) {
                        Participant sender = GetParticipant(ipAddress, endPoint.Port);
                        if (sender == null) {
                            sender = AddParticipant(ipAddress, endPoint.Port, this);
                            sender.isRemote = true;


                            UpdateEvent e = new() {
                                messageId = NetworkIdMsg.Id,
                                participant = sender
                            };

                            this.updateQueue.Enqueue(e);
                        }

                        ReceiveData(data, sender);
                    }
                }

                udpClient.BeginReceive(new AsyncCallback(callbackResult => ReceiveUDP(callbackResult)), null);
            }
            catch (Exception e) {
                Console.WriteLine($"Exception in communication thread: {e.Message}");
                Console.WriteLine($"Stack trace: {e.StackTrace}");
            }
        }

        public void ReceiveData(byte[] data, Participant sender) {
            int dataIx = 0;
            byte msgId = data[dataIx];
            if (msgId == 0xFF) {
                // Timeout
                return;
            }

            switch (msgId) {
                case ParticipantMsg.Id: { // 0xA0 / 160
                        ParticipantMsg msg = new(data);
                        this.Process(sender, msg);
                        dataIx += ParticipantMsg.length;
                    }
                    break;
                case NetworkIdMsg.Id: {// 0xA1 / 161
                        NetworkIdMsg msg = new(data);
                        this.Process(sender, msg);
                        dataIx += NetworkIdMsg.length;
                    }
                    break;
                case InvestigateMsg.Id: // 0x81
                    // result = await InvestigateMsg.Receive(dataStream, client, packetSize);
                    break;
                case ThingMsg.id: { // 0x80 / 128
                        ThingMsg msg = new(data);
                        this.Process(sender, msg);
                        dataIx += ThingMsg.length;
                    }
                    break;
                case NameMsg.Id: { // 0x91 / 145
                        NameMsg msg = new(data);
                        this.Process(sender, msg);
                        dataIx += NameMsg.length + msg.nameLength;
                    }
                    break;
                case ModelUrlMsg.Id: { // 0x90 / 144
                        ModelUrlMsg msg = new(data);
                        this.Process(sender, msg);
                        dataIx += ModelUrlMsg.length + msg.urlLength;
                    }
                    break;
                case PoseMsg.Id: { // 0x10 / 16
                        PoseMsg msg = new(data);
                        this.Process(sender, msg);
                        dataIx += msg.length;
                    }
                    break;
                case BinaryMsg.Id: { // 0xB1 / 177
                        BinaryMsg msg = new(data);
                        this.Process(sender, msg);
                        dataIx += BinaryMsg.length + msg.dataLength;
                    }
                    break;
                case TextMsg.Id: // 0xB0 / 176
                    // result = await TextMsg.Receive(dataStream, client, packetSize);
                    break;
                case DestroyMsg.Id: // 0x20 / 32
                    this.Process(sender, new DestroyMsg(data));
                    // result = await DestroyMsg.Receive(dataStream, client, packetSize);
                    break;
                default:
                    break;
            }
            if (dataIx < data.Length)
                Console.WriteLine($"####### Buffer not fully read, remaining {data.Length - dataIx}");
        }

        protected virtual void Process(Participant sender, ParticipantMsg msg) {
#if DEBUG
            Console.WriteLine($"{this.name} Process participantMsg {msg.networkId}");
#endif
        }

        protected virtual void Process(Participant sender, NetworkIdMsg msg) {
#if DEBUG            
            Console.WriteLine($"{this.name} Process SiteMsg {this.networkId} -> {msg.networkId}");
#endif

            if (this.networkId != msg.networkId) {
                this.networkId = msg.networkId;
                foreach (Thing thing in this.things) //Thing.GetAllThings())
                    sender.SendThingInfo(thing);

                // HACK to get the networkId for sites corrected to 0.
                // This is needed because AddParticipant assigns a networkId for non-sites
                sender.networkId = 0;
            }
        }

        protected virtual void Process(Participant sender, InvestigateMsg msg) {
#if DEBUG
            Console.WriteLine($"{this.name}: Process InvestigateMsg [{msg.networkId}/{msg.thingId}]");
#endif
        }

        protected virtual void Process(Participant sender, ThingMsg msg) {
#if DEBUG            
            Console.WriteLine($"{this.name}: Process ThingMsg [{msg.networkId}/{msg.thingId}] {msg.thingType} {msg.parentId}");
#endif            
            Participant owner = GetParticipant(msg.networkId);
            if (owner == null) {
                owner = new() {
                    networkId = msg.networkId
                };
                AddParticipant(owner);
                UpdateEvent e = new() {
                    messageId = ParticipantMsg.Id,
                    participant = owner
                };
                this.updateQueue.Enqueue(e);
                Console.WriteLine("Added Participant");
            }

            Thing thing = owner.Get(msg.networkId, msg.thingId);
            if (sender.networkId != owner.networkId) {
                // Sender forwarded this thing msg
                // So this is a remote thing
                thing ??= ProcessNewThing(owner, msg, true);
            }
            else {
                // Sender is the owner of the thing
                // So this is a simulated thing
                thing ??= ProcessNewThing(owner, msg, false);
            }

            if (msg.parentId != 0) {
                thing.parent = owner.Get(msg.networkId, msg.parentId);
                if (thing.parent == null)
                    Console.WriteLine($"Could not find parent [{msg.networkId}/{msg.parentId}]");
            }
            else {
                // Console.Write($"Dropped {thing.id}");
                thing.parent = null;
            }

            ForwardMessage(sender, msg);
        }

        protected virtual Thing ProcessNewThing(Participant owner, ThingMsg msg, bool isRemote) {
            Console.WriteLine("--- New Thing");
            Thing newThing = msg.thingType switch {
                Thing.Type.TouchSensor => new TouchSensor(owner.root),
                Thing.Type.DifferentialDrive => new DifferentialDrive(owner.root),
                _ => new Thing(owner.root)
            };
            newThing.id = msg.thingId;
            newThing.type = msg.thingType;
            newThing.isRemote = isRemote;
            return newThing;
        }

        protected virtual void Process(Participant sender, NameMsg msg) {
#if DEBUG            
            Console.WriteLine($"{this.name}: Process NameMsg [{msg.networkId}/{msg.thingId}] {msg.nameLength} {msg.name}");
#endif

            Participant owner = GetParticipant(msg.networkId);
            Thing thing = owner.Get(msg.networkId, msg.thingId);
            if (thing != null)
                thing.name = msg.name;

            ForwardMessage(sender, msg);
        }

        protected virtual void Process(Participant sender, ModelUrlMsg msg) {
#if DEBUG            
            Console.WriteLine($"{this.name}: Process ModelUrlMsg [{msg.networkId}/{msg.thingId}] {msg.urlLength} {msg.url}");
#endif

            Thing thing = sender.Get(msg.networkId, msg.thingId);
            if (thing != null)
                thing.modelUrl = msg.url;

            ForwardMessage(sender, msg);
        }

        protected virtual void Process(Participant sender, PoseMsg msg) {
#if DEBUG2         
            Console.WriteLine($"{this.name}: Process PoseMsg [{msg.networkId}/{msg.thingId}] {msg.poseType}");
#endif            
            Participant owner = Participant.GetParticipant(msg.networkId);
            if (owner == null)
                return;

            Thing thing = owner.Get(msg.networkId, msg.thingId);
            if (thing == null)
                return;

            if ((msg.poseType & PoseMsg.Pose_Position) != 0)
                thing.position = msg.position;
            if ((msg.poseType & PoseMsg.Pose_Orientation) != 0)
                thing.orientation = msg.orientation;
            if ((msg.poseType & PoseMsg.Pose_LinearVelocity) != 0)
                thing.linearVelocity = msg.linearVelocity;
            if ((msg.poseType & PoseMsg.Pose_AngularVelocity) != 0)
                thing.angularVelocity = msg.angularVelocity;
        }

        protected virtual void Process(Participant sender, BinaryMsg msg) {
#if DEBUG
            //            Console.WriteLine($"{this.name}: Process BinaryMsg [{msg.networkId}/{msg.thingId}] {msg.dataLength}");
#endif
            Thing thing = sender.Get(msg.networkId, msg.thingId);
            thing?.ProcessBinary(msg.data);
        }

        protected virtual void Process(Participant sender, TextMsg msg) {
#if DEBUG
            Console.WriteLine($"Participant: Process TextMsg {msg.textLength} {msg.text}");
#endif            

        }

        protected virtual void Process(Participant sender, DestroyMsg msg) {
#if DEBUG
            Console.WriteLine($"Participant: Process Destroy Msg [{msg.networkId}/{msg.thingId}]");
#endif            
            Thing thing = sender.Get(msg.networkId, msg.thingId);
            if (thing != null)
                this.Remove(thing);
#if UNITY_5_3_OR_NEWER
            thing.component.core = null;
#endif
        }

        private void ForwardMessage(Participant sender, IMessage msg) {
            foreach (Participant participant in Participant.participants) {
                if (participant == this || participant == sender)
                    continue;

                // Console.WriteLine($"---> {participant.networkId}");
                participant.Send(msg);
            }
        }

        #endregion

        #region Participant Registry

        /// <summary>
        /// Retrieve a participant using ip address and port number
        /// </summary>
        /// <param name="ipAddress">The ip address of the participant</param>
        /// <param name="port">The port number used to send messages to the participant</param>
        /// <returns>The participant or null if it is not found.</returns>
        public static ParticipantUDP GetParticipant(string ipAddress, int port) {
            //Console.WriteLine($"Get Participant {ipAddress}:{port}");
            foreach (Participant participant in Participant.participants) {
                ParticipantUDP participantUDP = participant as ParticipantUDP;
                if (participantUDP.ipAddress == ipAddress && participantUDP.port == port)
                    return participantUDP;
            }
            return null;
        }

        /// <summary>
        /// Add a new participant to the collection of participants
        /// </summary>
        /// <param name="ipAddress">The IP address of the participant</param>
        /// <param name="port">The port used to send messages to this participant</param>
        /// <returns>The added participant</returns>
        public static ParticipantUDP AddParticipant(string ipAddress, int port, ParticipantUDP localParticipant = null) {
            Console.WriteLine($"New Participant {ipAddress}:{port}");
            // This code is only valid for site, because those can distribute networkIds.....
            ParticipantUDP participant = new(ipAddress, port, localParticipant) {
                networkId = (byte)(Participant.participants.Count + 1)
            };
            Participant.participants.Add(participant);
            return participant;
        }
        #endregion
    }
}